import json
from collections import OrderedDict, namedtuple


primitive = (str, int, float, bool, type(None))



class MissingEntryError(TypeError):
	def __init__(self, name=None, cls=None, ancestors=None):
		msg = []
		if name is not None:
			msg.append(f'name={name}')
		if cls is not None:
			msg.append(f'cls={cls}')
		if ancestors is not None:
			msg.append(f'ancestors={ancestors}')
		msg = ', '.join(msg)
		super().__init__(msg)




class Registry(OrderedDict):
	
	_pack_fn_name = '__pack__'
	_unpack_fn_name = '__unpack__'
	_create_fn_name = '__create__'
	
	def __init__(self, primary_component='name', sister_component='cls',
	             components=['pack_fn', 'create_fn', 'unpack_fn', 'ancestors'],
	             sister_registry_object=None, sister_registry_cls=None):
		if sister_registry_cls is None:
			sister_registry_cls = self.__class__
		if sister_registry_object is None:
			sister_registry_object = sister_registry_cls(sister_registry_object=self,
			                                             primary_component=sister_component,
			                                             sister_component=primary_component,
			                                             components=components)
		
		super().__init__()
		self._sister_registry_object = sister_registry_object
		
		self._key_name = primary_component
		self._sister_key_name = sister_component
		self._entry_cls = namedtuple(f'{self.__class__.__name__}_Entry',
		                             [primary_component, sister_component] + components)
		
		self._init_sister_registry()
	
	
	def is_known(self, x):
		return x in self or x in self.backward()
	
	
	def backward(self):
		return self._sister_registry_object
	
	
	def _init_sister_registry(self):
		for k, v in self.items():
			self._sister_registry_object.__setitem__(self._get_sister_entry_key(k, v), v, sync=False)
	
	
	def _get_sister_entry_key(self, key, value):
		assert isinstance(value, self._entry_cls)
		if key == getattr(value, self._key_name):
			return getattr(value, self._sister_key_name)
		return getattr(value, self._key_name)
	
	
	def update(self, other, sync=True):
		if sync:
			self._sister_registry_object.update({self._get_sister_entry_key(k, v): v
			                                     for k, v in other.items()}, sync=False)
		return super().update(other)


	def __setitem__(self, key, value, sync=True):
		if sync:
			self._sister_registry_object.__setitem__(self._get_sister_entry_key(key, value), value, sync=False)
		return super().__setitem__(key, value)
	
	
	def __delitem__(self, key, sync=True):
		if sync:
			self._sister_registry_object.__delitem__(self._get_sister_entry_key(key, self[key]), sync=False)
		return super().__delitem__(key)
		
		
	def _old_new(self, name, obj):  # register a new entry
		self[name] = obj
		return obj


	def _old_new2(self, *args, **info):  # register a new entry
		if self._key_name not in info:
			assert len(args) == 1
			info[self._key_name] = args[0]
		assert self._key_name in info, f'Missing key: {self._key_name}'
		return self._old_new(info[self._key_name], self._entry_cls(**info))
	
	
	def is_registered(self, obj):
		for opt in self.values():
			if obj == opt:
				return True
		return False
	
	
	def new(self, cls, name=None, pack_fn=None, create_fn=None, unpack_fn=None, ancestors=None):
		
		if name is None:
			name = self.full_name(cls)
		
		if pack_fn is None:
			pack_fn = getattr(cls, self._pack_fn_name, None)
		
		if create_fn is None:
			create_fn = getattr(cls, self._create_fn_name, None)
			if create_fn is not None:
				create_fn = create_fn.__func__
		
		if unpack_fn is None:
			unpack_fn = getattr(cls, self._unpack_fn_name, None)
		
		if ancestors is None:
			ancestors = self._find_ancestors(cls)
		ancestors = [(parent if parent in self else self.backward()[parent].name)
		             for parent in ancestors if self.is_known(parent)]
		
		return self._old_new2(name=name, cls=cls, ancestors=ancestors,
		                   pack_fn=pack_fn, create_fn=create_fn, unpack_fn=unpack_fn)
	
	
	def find_nearest(self, name=None, cls=None, ancestors=None):
		if name is not None and name in self:
			return self[name]
		if cls is not None and cls in self.backward():
			return self.backward()[cls]
		
		if ancestors is not None:
			for ancestor in ancestors:
				if ancestor in self:
					return self[ancestor]
		
		raise MissingEntryError(name=name, cls=cls, ancestors=ancestors)
	
	
	def _find_ancestors(self, cls):
		return cls.__mro__[1:]
	
	
	@staticmethod
	def full_name(cls):
		name = str(cls.__name__)
		module = str(cls.__module__)
		if module is None:
			return name
		return '.'.join([module, name])



registry = Registry()
register_packable = registry.new

def as_packable(name=None, ancestors=None, pack_fn=None, create_fn=None, unpack_fn=None):
	def _reg_pack(cls):
		register_packable(name=name, cls=cls, ancestors=ancestors,
                pack_fn=pack_fn, create_fn=create_fn, unpack_fn=unpack_fn)
		return cls
	return _reg_pack


for _primitive in primitive:
	register_packable(_primitive, create_fn=lambda _, d, __, ___: d)


# region Mutable

def _pack_dict(d, pack_member):
	return {pack_member(k,force_str=True):pack_member(v) for k,v in d.items()}
def _unpack_dict(d, x, o, unpack_member):
	d.update({unpack_member(k):unpack_member(v) for k, v in x.items()})
register_packable(dict, ancestors=[],
             pack_fn=_pack_dict,
             unpack_fn=_unpack_dict,
)
register_packable(OrderedDict,
             pack_fn=_pack_dict,
             unpack_fn=_unpack_dict,
)

def _pack_list(xs, pack_member):
	return [pack_member(x) for x in xs]
def _unpack_list(ls, xs, o, unpack_member):
	ls.extend(unpack_member(x) for x in xs)
register_packable(list, ancestors=[],
             pack_fn=_pack_list,
             unpack_fn=_unpack_list,
)
register_packable(set, ancestors=[list],
             pack_fn=_pack_list,
             unpack_fn=lambda s, xs, o, unpack_member: s.update(unpack_member(x) for x in xs),
)

# endregion

# region Immutable

class TypeStr(str):
	pass

register_packable(type, name='class', ancestors=[],
             pack_fn=lambda t, _: registry.backward()[t].name if t in registry.backward() else registry.full_name(t),
             create_fn=lambda _, d, __, ___: registry[d].cls if d in registry else TypeStr(d),
                  # TODO check ancestry for nearest
)
register_packable(tuple, ancestors=[list],
             pack_fn=_pack_list,
             create_fn=lambda _, xs, __, ___: tuple(unpack_member(x) for x in xs),
)
register_packable(complex, ancestors=[dict],
             pack_fn=lambda c, _: [pack_member(c.real), pack_member(c.imag)],
             create_fn=lambda _, data, __, ___: complex(unpack_member(data[0]), unpack_member(data[1])),
)
register_packable(range, ancestors=[dict],
             pack_fn=lambda r, _: {'start':pack_member(r.start), 'stop':pack_member(r.stop), 'step':pack_member(r.step)},
             create_fn=lambda _, data, __, ___: range(start=unpack_member(data['start']), stop=unpack_member(data['stop']),
                                          step=unpack_member(data['step'])),
)
register_packable(bytes, ancestors=[str],
             pack_fn=lambda b, _: b.decode(),
             create_fn=lambda _, s, __, ___: s.encode('latin1'),
)

# endregion


class Packable(object):
	
	def __init_subclass__(cls, use_cls = None, name = None):
		super().__init_subclass__()  # TODO: remove
		
		if use_cls is None:
			use_cls = cls
		
		register_packable(use_cls, name=name)
	
	def __deepcopy__(self, memodict = None):
		return unpack(pack(self))
	
	
	@classmethod
	def __create__(cls, data, original, unpack_member):
		return cls.__new__(cls)
	
	
	def __pack__(self, pack_member):
		return {pack_member(k, force_str=True): pack_member(v) for k, v in self.__dict__.items()}
	
	
	def __unpack__(self, data, original, unpack_member):
		self.__dict__.update({unpack_member(k): unpack_member(v) for k, v in data.items()})


class Packer:
	def __init__(self, key_prefix='!', ref_prefix='$'):
		self.key_prefix = key_prefix
		self.ref_prefix = ref_prefix
		self.reset()
	
	def reset(self):
		self.ref_table = {}
		self.obj_table = {}
		self.ancestry = {}
		self.corrections = []
	
	def make_obj_ref(self, obj):
		return f'{self.ref_prefix}{hex(id(obj))[2:]}'
	
	def find_nearest(self, name=None, cls=None, ancestors=None):
		return registry.find_nearest(name=name, cls=cls, ancestors=ancestors)
	
	def pack_member(self, obj, force_str = False):
		if isinstance(obj, primitive):
			if (isinstance(obj, str) and obj.startswith(self.ref_prefix)) or (not isinstance(obj, str) and force_str):
				ref = self.make_obj_ref(obj)
				self.ref_table[ref] = {f'{self.key_prefix}type': self.find_nearest(cls=type(obj)).name,
				                       f'{self.key_prefix}data': obj}
			else:
				return obj
		else:
			ref = self.make_obj_ref(obj)
			typ = type(obj)
			
			if ref in self.ref_table:
				return ref
			data = {}
			self.ref_table[ref] = data  # create entry in refs to stop reference loops
			
			info = self.find_nearest(cls=typ)
			self.ancestry[info.name] = info.ancestors
			data[f'{self.key_prefix}type'] = info.name
			data[f'{self.key_prefix}data'] = self._dispatch_pack(info, obj)
		
		return ref
	
	def unpack_member(self, data):
		
		if isinstance(data, str) and data.startswith(self.ref_prefix):  # reference or class
			if data in self.obj_table:  # known reference
				return self.obj_table[data]
			
			ref = data
			typname = self.ref_table[ref][f'{self.key_prefix}type']
			data = self.ref_table[ref][f'{self.key_prefix}data']
			
			info = self.find_nearest(name=typname, ancestors=self.ancestry.get(typname, None))
			
			obj = self._dispatch_create(info, data)
			
			del self.ref_table[ref]
			self.obj_table[ref] = obj
			
			self._dispatch_unpack(info, obj, data)
			
			if self.corrections is not None and info.name != typname:
				self.corrections.append({'expected': typname, 'used': info.name, 'obj': obj})
		
		else:
			
			info = self.find_nearest(cls=type(data))
			
			obj = self._dispatch_create(info, data)
			self._dispatch_unpack(info, obj, data)
			
		return obj
	
	def _dispatch_pack(self, entry, obj):
		return obj if entry.pack_fn is None else entry.pack_fn(obj, self.pack_member)
	
	def _dispatch_create(self, entry, data, original=None):
		if entry.create_fn is None and entry.unpack_fn is None:
			return entry.cls(data)
		elif entry.create_fn is None:
			return entry.cls()
		return entry.create_fn(entry.cls, data, original, self.unpack_member)
	
	def _dispatch_unpack(self, entry, obj, data, original=None):
		if entry.unpack_fn is not None:
			return entry.unpack_fn(obj, data, original, self.unpack_member)
	
	def pack(self, obj, meta = None, include_timestamp = False):
		self.reset()
		
		try:
			out = self.pack_member(obj)
			
			# additional meta info
			if meta is None:
				meta = {}
			if include_timestamp:
				meta['timestamp'] = time.strftime('%Y-%m-%d_%H%M%S')
			
			data = {
				'table': self.ref_table,
				'ancestry': self.ancestry,
				'meta': meta,
				'head': out,  # save parent object separately
			}
		
		except Exception as e:
			raise e
		
		return data
	
	def unpack(self, data, return_meta = False, return_corrections = False, allow_ancestors = True):
		self.reset()
		self.ref_table = data.get('table', {})
		self.ancestry = data.get('ancestry', {}) if allow_ancestors else {}
		
		try:
			obj = self.unpack_member(data['head'])
		except Exception as e:
			raise e
		
		out = [obj]
		
		if allow_ancestors and return_corrections:
			out.append(self.corrections)
		if return_meta:
			out.append(data.get('meta', None))
		
		if len(out) > 1:
			return out
		return out[0]



def pack(obj, meta=None, include_timestamp=False, packer=None):
	if packer is None:
		packer = Packer()
	return packer.pack(obj, meta=meta, include_timestamp=include_timestamp)



def unpack(data, return_meta=False, return_corrections=False,
           allow_ancestors=True, packer=None, skip_deepcopy=False):
	if packer is None:
		packer = Packer()
	if not skip_deepcopy:
		data = json.loads(json.dumps(data))
	return packer.unpack(data, return_meta=return_meta, return_corrections=return_corrections)


def json_pack(obj, meta=None, include_timestamp=False):
	return json.dumps(pack(obj, meta=meta, include_timestamp=include_timestamp))


def json_unpack(data, return_meta=False, return_corrections=False, allow_ancestors=True):
	return unpack(json.loads(data), return_meta=return_meta, return_corrections=return_corrections,
	              allow_ancestors=allow_ancestors)



	




from typing import Any
from wrapt import ObjectProxy


from .packing import Packable, register_packable
from .transactions import Transactionable
from .basic_containers import adict, tset, tlist


# all wrapped objects must be able to be copied (shallow copy) using
# note: Transactionable objects cant be wrapped
class ObjectWrapper(Transactionable, Packable, ObjectProxy):
	'''
	Wrapper to transform an object to be transactionable.

	Note: wrapped object must be copyable (shallow copy using `.copy()`)

	WARNING: It is NOT recommended to use this wrapper, unless you need a transactionable features

	'''
	
	def __init__(self, obj):
		super().__init__(obj)
		
		self._self_shadow = None
		self._self_children = tset()
	
	def begin(self):
		if self.in_transaction():
			return
			self.commit()
		
		self._self_shadow = self.copy()
		self._self_children.begin()
	
	def in_transaction(self):
		return self._self_shadow is not None
	
	def commit(self):
		if not self.in_transaction():
			return
		
		self._self_shadow = None
		self._self_children.commit()
	
	def abort(self):
		if not self.in_transaction():
			return
		
		self.__wrapped__ = self._self_shadow
		self._self_shadow = None
		self._self_children.abort()
	
	def __repr__(self):
		return self.__wrapped__.__repr__()
	
	def __str__(self):
		return self.__wrapped__.__str__()
	
	def __setattr__(self, key, value):
		if isinstance(value, Transactionable) and not key == '_self_children':
			self._self_children.add(value)
		return super().__setattr__(key, value)
	
	def __delattr__(self, item):
		value = self.__getattr__(item)
		if isinstance(value, Transactionable):
			self._self_children.remove(value)
		return super().__delattr__(item)
	
	def __unpack__(self, data, o, unpack_member):
		obj = self.__build__(data, o, unpack_member)
		
		self.__init__(obj)
	
	# must be overridden
	
	def __pack__(self, pack_member):  # save everything from the internal state
		'''
		Save all the necessary data from the internal state (and pack any subdata)

		:return: packed data
		'''
		raise NotImplementedError
	
	def __build__(self, data, o, unpack_member):
		'''
		Recover the wrapped object in the correct state from data and return wrapped object

		:param data: packed data
		:return: wrapped object with the loaded state
		'''
		raise NotImplementedError


try:
	import numpy as np
	
	#
	# class Packable_Array(Packable, use_cls=np.ndarray):
	# 	'''
	# 	Wrapper to allow saving numpy arrays.
	# 	Aside from being rather useful, this serves as an example for how to write a Packable wrapper.
	#
	# 	Note the necessary Packable methods are all static, and the use of "use_cls" in the class declaration.
	# 	'''
	#
	# 	@staticmethod
	# 	def __create__(data):
	# 		'''
	# 		Creates an empty np.array
	#
	# 		:param data: packed data
	# 		:return: empty array with the correct size
	# 		'''
	# 		shape, dtype = data['shape'], data['dtype']
	# 		return np.empty(shape, dtype)
	#
	# 	@staticmethod
	# 	def __pack__(obj, pack_member):
	# 		'''
	# 		Pack the np.array data.
	#
	# 		Note: that the information necessary for creating thet instance (shape, dtype) is not packed,
	# 		but still valid json objects
	#
	# 		:param obj: instance of numpy.ndarray to be packed
	# 		:return: packed data
	# 		'''
	#
	# 		data = {}
	#
	# 		data['shape'] = list(obj.shape)
	# 		data['dtype'] = obj.dtype.name
	#
	# 		data['data'] = pack_member(obj.tolist())
	#
	# 		return data
	#
	# 	@staticmethod
	# 	def __unpack__(obj, data, unpack_member):
	# 		'''
	# 		Unpack the data and save the data to the created object
	#
	# 		:param obj: instance with empty data to populate with the unpacked data
	# 		:param data: packed data
	# 		:return: None
	# 		'''
	#
	# 		obj[:] = np.array(unpack_member(data['data']), dtype=data['dtype'])
	#
	#
	# class Array(ObjectWrapper):
	# 	'''
	# 	This is an example of how to use the `ObjectWrapper`.
	# 	Wraps numpy arrays.
	#
	# 	WARNING: it is NOT recommended to use this wrapper for numpy arrays (they are already registered).
	# 	'''
	#
	# 	def begin(self):
	# 		super().begin()
	# 		if self.dtype.name == 'object':
	# 			for el in self.flat:
	# 				if isinstance(el, Transactionable):
	# 					el.begin()
	#
	# 	def commit(self):
	# 		super().commit()
	# 		if self.dtype.name == 'object':
	# 			for el in self.flat:
	# 				if isinstance(el, Transactionable):
	# 					el.commit()
	#
	# 	def abort(self):
	# 		super().abort()
	# 		if self.dtype.name == 'object':
	# 			for el in self.flat:
	# 				if isinstance(el, Transactionable):
	# 					el.abort()
	#
	# 	def __pack__(self, pack_member):
	# 		'''
	# 		Pack data to restore numpy array.
	#
	# 		:return: packed data
	# 		'''
	# 		data = {
	# 			'dtype': pack_member(self.dtype.name),
	# 			'data': pack_member(self.tolist()),
	# 			'shape': pack_member(self.shape),
	# 		}
	#
	# 		return data
	#
	# 	def __build__(self, data, o, unpack_member):
	# 		'''
	# 		Restore state of numpy array by unpacking data
	#
	# 		:param data: packed data
	# 		:return: restored state
	# 		'''
	# 		return np.array(unpack_member(data['data']), dtype=unpack_member(data['dtype']))
	#
	
	def _unpack_array(arr, data, _, unpack_member):
		arr[:] = np.array(unpack_member(data['data']), dtype=unpack_member(data['dtype']))
	
	register_packable(np.ndarray,
	                  pack_fn=lambda arr, pack_member: {'shape':pack_member(arr.shape),
	                                                    'dtype':pack_member(str(arr.dtype)),
	                                                    'data': pack_member(arr.tolist())},
	                  create_fn=lambda _, data, __, unpack_member: np.empty(unpack_member(data['shape']),
	                                                                        unpack_member(data['dtype'])),
	                  unpack_fn=_unpack_array,
	                  )


except ImportError:
	print('WARNING: unable to import numpy')







from typing import Any, Union, Dict, List, Set, Tuple, NoReturn, ClassVar, TextIO, Callable, NewType
import json
from pathlib import Path
from collections import namedtuple, OrderedDict
import time

import copy

from .registries import Entry_Double_Registry
# from .errors import SavableClassCollisionError, ObjectIDReadOnlyError, UnregisteredClassError
from .logging import get_printer

prt = get_printer(__file__)

primitive = (str, int, float, bool, type(None)) # all json readable and no sub elements

# py_types = (bytes, complex, range, tuple)
# py_containers = (dict, list, set)


class MissingEntryError(TypeError):
	def __init__(self, name=None, cls=None, ancestors=None):
		msg = []
		if name is not None:
			msg.append(f'name={name}')
		if cls is not None:
			msg.append(f'cls={cls}')
		if ancestors is not None:
			msg.append(f'ancestors={ancestors}')
		msg = ', '.join(msg)
		super().__init__(msg)



class Packable_Registry(Entry_Double_Registry, primary_component='name', sister_component='cls',
                        components=['pack_fn', 'create_fn', 'unpack_fn', 'ancestors']):
	
	_pack_fn_name = '__pack__'
	_unpack_fn_name = '__unpack__'
	_create_fn_name = '__create__'
	
	
	def new(self, cls, name=None, pack_fn=None, create_fn=None, unpack_fn=None, ancestors=None):
		''' TODO: update
		Register a type to be packable. Requires a pack_fn, create_fn, and unpack_fn to store and restore object state.

		:param cls: type to be registered
		:param pack_fn: callable input is an instance of the type, and packs all data necessary to recover the state
		:param create_fn: callable input is the expected type and the packed data, creates a new instance of the type,
		without unpacking any packed data (to avoid reference loops)
		:param unpack_fn: callable input is the instance of packed data and then restores that instance to the original
		state using the packed data by unpacking any values therein.
		:param name: (optional) name of the class used for storing
		:return: A `SavableClassCollisionError` if the name is already registered
		'''
		
		if name is None:
			name = self.full_name(cls)
			
		if pack_fn is None:
			pack_fn = getattr(cls, self._pack_fn_name, None)
		
		if create_fn is None:
			create_fn = getattr(cls, self._create_fn_name, None)
			if create_fn is not None:
				create_fn = create_fn.__func__
		
		if unpack_fn is None:
			unpack_fn = getattr(cls, self._unpack_fn_name, None)
		
		if ancestors is None:
			ancestors = self._find_ancestors(cls)
		ancestors = [(parent if parent in self else self.backward()[parent].name)
		             for parent in ancestors if self.is_known(parent)]
		
		return super().new(name=name, cls=cls, ancestors=ancestors,
		                   pack_fn=pack_fn, create_fn=create_fn, unpack_fn=unpack_fn)
	
	
	def find_nearest(self, name=None, cls=None, ancestors=None):
		if name is not None and name in self:
			return self[name]
		if cls is not None and cls in self.backward():
			return self.backward()[cls]
		
		if ancestors is not None:
			for ancestor in ancestors:
				if ancestor in self:
					return self[ancestor]
		
		raise MissingEntryError(name=name, cls=cls, ancestors=ancestors)
	
	
	def _find_ancestors(self, cls):
		return cls.__mro__[1:]
	
	
	@staticmethod
	def full_name(cls: ClassVar) -> str:
		'''
		Find the full, unique name of a class by connecting it to the module where is is declared.

		:param cls: type
		:return: unique name of the class
		'''
		name = str(cls.__name__)
		module = str(cls.__module__)
		if module is None:
			return name
		return '.'.join([module, name])



registry = Packable_Registry()
register_packable = registry.new

def as_packable(name=None, ancestors=None, pack_fn=None, create_fn=None, unpack_fn=None):
	def _reg_pack(cls):
		register_packable(name=name, cls=cls, ancestors=ancestors,
                pack_fn=pack_fn, create_fn=create_fn, unpack_fn=unpack_fn)
		return cls
	return _reg_pack


for _primitive in primitive:
	register_packable(_primitive, create_fn=lambda _, d, __, ___: d)


# region Mutable

def _pack_dict(d, pack_member):
	return {pack_member(k,force_str=True):pack_member(v) for k,v in d.items()}
def _unpack_dict(d, x, o, unpack_member):
	d.update({unpack_member(k):unpack_member(v) for k, v in x.items()})
register_packable(dict, ancestors=[],
             pack_fn=_pack_dict,
             unpack_fn=_unpack_dict,
)
register_packable(OrderedDict,
             pack_fn=_pack_dict,
             unpack_fn=_unpack_dict,
)

def _pack_list(xs, pack_member):
	return [pack_member(x) for x in xs]
def _unpack_list(ls, xs, o, unpack_member):
	ls.extend(unpack_member(x) for x in xs)
register_packable(list, ancestors=[],
             pack_fn=_pack_list,
             unpack_fn=_unpack_list,
)
register_packable(set, ancestors=[list],
             pack_fn=_pack_list,
             unpack_fn=lambda s, xs, _, unpack_member: s.update(unpack_member(x) for x in xs),
)

# endregion

# region Immutable

register_packable(type, name='class', ancestors=[],
             pack_fn=lambda t, _: registry.backward()[t].name if t in registry.backward() else registry.full_name(t),
             create_fn=lambda _, d, __, ___: registry[d].cls if d in registry else locals().get(d),
                  # TODO check ancestry for nearest
)
register_packable(tuple, ancestors=[list],
             pack_fn=_pack_list,
             create_fn=lambda _, xs, __, unpack_member: tuple(unpack_member(x) for x in xs),
)
register_packable(complex, ancestors=[dict],
             pack_fn=lambda c, pack_member: [pack_member(c.real), pack_member(c.imag)],
             create_fn=lambda _, data, __, unpack_member: complex(unpack_member(data[0]), unpack_member(data[1])),
)
register_packable(range, ancestors=[dict],
             pack_fn=lambda r, pack_member: {'start':pack_member(r.start),
                                             'stop':pack_member(r.stop),
                                             'step':pack_member(r.step)},
             create_fn=lambda _, data, __, unpack_member: range(unpack_member(data['start']),
                                                                unpack_member(data['stop']),
                                                                unpack_member(data['step'])),
)
register_packable(bytes, ancestors=[str],
             pack_fn=lambda b, _: b.decode('8859'),
             create_fn=lambda _, s, __, ___: s.encode('8859'),
)

# endregion


class Packable(object):
	'''
	Any subclass of this mixin can be serialized using `pack`
	
	All subclasses must implement __create__, __pack__, and __unpack__ to register the type. By passing a type to
	`use_cls` the type for which these methods are used can be overridden from the subclass.
	'''
	def __init_subclass__(cls, use_cls: ClassVar = None, name: str = None) -> NoReturn:
		'''
		This method automatically registers any subclass that is declared.
		
		:param use_cls: The class to register (if it is different than `cls`)
		:return: None
		'''
		super().__init_subclass__() # TODO: remove
		
		if use_cls is None:
			use_cls = cls
		
		register_packable(use_cls, name=name)


	def __deepcopy__(self, memodict: Dict[Any,Any] = None) -> Any:
		'''
		Produces a deep copy of the data by packing and repacking.
		
		:param memodict: Unused
		:return: A deep copy of self
		'''
		return unpack(pack(self))
	
	
	@classmethod
	def __create__(cls, data: Dict[str, 'PACKED'], original: str, unpack_member: Callable) -> 'Packable':
		'''
		Create the object without loading the state from data. You can use the data to inform how
		to initialize the object, however no stored objects should be unpacked (to avoid reference loops)
		
		:param data: packed data to restore object state, should NOT be unpacked here
		:return: A fresh instance of the class registered with this create_fn
		'''
		return cls.__new__(cls)
	
	
	def __pack__(self, pack_member: Callable) -> Dict[str,'PACKED']:
		'''
		Collect all data in self necessary to store the state.
		
		.. warning:: All data must be "packed" storing it. This is done by passing the data into
		`Packable._pack_obj` and using what is returned.
		
		:return: A dict of packed data necessary to recover the state of self
		'''
		return {pack_member(k, force_str=True): pack_member(v) for k,v in self.__dict__.items()}
	
	
	def __unpack__(self, data: Dict[str, 'PACKED'], original: str, unpack_member: Callable) -> NoReturn:
		'''
		Using `data`, recover the packed state.
		Must be overridden by all subclasses.
		
		.. warning:: All data must be "unpacked" before using it. This is done by passing the data into
		`Packable._unpack_obj` and using what is returned.
		
		:param data: The information that is returned by `__pack__`.
		:return: Nothing. Once returned, the object should be in the same state as when it was packed
		'''
		self.__dict__.update({unpack_member(k): unpack_member(v) for k, v in data.items()})



Primitive = Union[primitive]
'''Valid primitives'''
#
# SERIALIZABLE = Union[Packable, PRIMITIVE, Dict['SERIALIZABLE', 'SERIALIZABLE'],
#                      List['SERIALIZABLE'], Set['SERIALIZABLE'], Tuple['SERIALIZABLE']]
# '''Types that can be serialized using `pack`'''
#
JSONABLE = Union[Dict[str,'JSONABLE'], List['JSONABLE'], Primitive]
'''Any object that is valid in json (eg. using `json.dumps`)'''
#
# PACKED = Union[PRIMITIVE, List['PACKED'], Dict['PACKED', 'PACKED']]
# '''Any information that is valid json and can be unpacked to recover the state of `Packable` subclasses.'''



SERIALIZABLE = NewType('SERIALIZABLE', object)
# JSONABLE = NewType('JSONABLE', object)
PACKED = NewType('PACKED', object)



class Packer:
	def __init__(self, key_prefix='!', ref_prefix='$'):
		self.key_prefix = key_prefix
		self.ref_prefix = ref_prefix
		self.reset()
		
		
	def reset(self):
		self.ref_table = {}
		self.obj_table = {}
		self.ancestry = {}
		self.corrections = []
	
	
	def make_obj_ref(self, obj: SERIALIZABLE) -> str:
		'''
		Compute the object ID for packing objects, which must be unique and use the reference prefix

		:param obj: object to get the reference for
		:return: unique ID associated with `obj` for packing
		'''
		return f'{self.ref_prefix}{hex(id(obj))[2:]}'
	
	
	def find_nearest(self, name=None, cls=None, ancestors=None):
		return registry.find_nearest(name=name, cls=cls, ancestors=ancestors)
		
	
	def pack_member(self, obj: SERIALIZABLE, force_str: bool = False) -> PACKED:
		'''
		Store the object state by packing it, possibly returning a reference.
		
		This function should be called inside implemented __pack__ on all data in an object necessary to restore
		the object state.
		
		Note: this function should not be called on the top level (use `pack` instead).
		
		:param obj: serializable data that should be packed
		:param force_str: if the data is a key for a dict, set this to true to ensure the key is a str
		:return: packed data
		'''
		if isinstance(obj, primitive):
			if (isinstance(obj, str) and obj.startswith(self.ref_prefix)) or (not isinstance(obj, str) and force_str):
				ref = self.make_obj_ref(obj)
				self.ref_table[ref] = {f'{self.key_prefix}type': self.find_nearest(cls=type(obj)).name,
				                       f'{self.key_prefix}data': obj}
			else:
				return obj
		else:
			ref = self.make_obj_ref(obj)
			typ = type(obj)
	
			if ref in self.ref_table:
				return ref
			data = {}
			self.ref_table[ref] = data  # create entry in refs to stop reference loops
			
			info = self.find_nearest(cls=typ)
			self.ancestry[info.name] = info.ancestors
			data[f'{self.key_prefix}type'] = info.name
			data[f'{self.key_prefix}data'] = self._dispatch_pack(info, obj)
		
		return ref
	
	
	def unpack_member(self, data: PACKED) -> SERIALIZABLE:
		'''
		Restore the object data by unpacking it.
		
		This function should be called inside implemented __unpack__ on all data in an object necessary to restore
		the object state from the packed data.
		
		Note: this function should not be called on the top level (use `unpack` instead).
		
		:param data: packed data that should be unpacked
		:return: unpacked data to restore the state
		'''
	
		if isinstance(data, str) and data.startswith(self.ref_prefix):  # reference or class
			if data in self.obj_table: # known reference
				return self.obj_table[data]
	
			ref = data
			typname = self.ref_table[ref][f'{self.key_prefix}type']
			data = self.ref_table[ref][f'{self.key_prefix}data']
			
			info = self.find_nearest(name=typname, ancestors=self.ancestry.get(typname, None))
			
			obj = self._dispatch_create(info, data, typname)
			
			del self.ref_table[ref]
			self.obj_table[ref] = obj
			
			self._dispatch_unpack(info, obj, data, typname)
			
			if self.corrections is not None and info.name != typname:
				self.corrections.append({'expected': typname, 'used': info.name, 'obj': obj})
			
		else:
			info = self.find_nearest(cls=type(data))
			
			obj = self._dispatch_create(info, data)
			self._dispatch_unpack(info, obj, data)
			
		
		return obj


	def _dispatch_pack(self, entry, obj):
		return obj if entry.pack_fn is None else entry.pack_fn(obj, self.pack_member)
	
	
	def _dispatch_create(self, entry, data, original=None):
		if entry.create_fn is None and entry.unpack_fn is None:
			return entry.cls(data)
		elif entry.create_fn is None:
			return entry.cls()
		return entry.create_fn(entry.cls, data, original, self.unpack_member)
	
	
	def _dispatch_unpack(self, entry, obj, data, original=None):
		if entry.unpack_fn is not None:
			return entry.unpack_fn(obj, data, original, self.unpack_member)

	
	def pack(self, obj: SERIALIZABLE, meta: Dict[str, PACKED] = None, include_timestamp: bool = False) -> JSONABLE:
		'''
		Serializes any object, returning a json object that can be converted to a json string.
		
		:param obj: Object to be serialized
		:param meta: Meta information, must be jsonable
		:param include_timestamp: include a timestamp in the meta information
		:return: packed data, which can be converted to a json string using json.dumps
		'''
		self.reset()
	
		try:
			out = self.pack_member(obj)
	
			# additional meta info
			if meta is None:
				meta = {}
			if include_timestamp:
				meta['timestamp'] = time.strftime('%Y-%m-%d_%H%M%S')
	
			data = {
				'table': self.ref_table,
				'ancestry': self.ancestry,
				'meta': meta,
				'head': out, # save parent object separately
			}
	
		except Exception as e:
			raise e

		return data
	
	
	def unpack(self, data: PACKED, return_meta: bool = False, return_corrections: bool = False,
	           allow_ancestors: bool = True) -> SERIALIZABLE:
		'''
		Deserialize a packed object to recover the original state.
		
		:param data: serialized (packed) state of an object
		:param return_meta: return any meta information from the serialized data
		:return: the unpacked (restored) object
		'''
		self.reset()
		self.ref_table = data.get('table', {})
		self.ancestry = data.get('ancestry', {}) if allow_ancestors else {}
	
		try:
			obj = self.unpack_member(data['head'])
		except Exception as e:
			raise e
	
		out = [obj]
		
		if allow_ancestors and return_corrections:
			out.append(self.corrections)
		if return_meta:
			out.append(data.get('meta', None))
	
		if len(out) > 1:
			return out
		return out[0]



def pack(obj: SERIALIZABLE, meta: Dict[str, PACKED] = None, include_timestamp: bool = False,
         packer: Union[Packer, None] = None) -> JSONABLE:
	if packer is None:
		packer = Packer()
	return packer.pack(obj, meta=meta, include_timestamp=include_timestamp)
	
	
	
def unpack(data: PACKED, return_meta: bool = False, return_corrections: bool = False,
           allow_ancestors: bool = True, packer: Union[Packer, None] = None,
           skip_deepcopy: bool = False) -> SERIALIZABLE:
	if packer is None:
		packer = Packer()
	if not skip_deepcopy:
		data = json.loads(json.dumps(data))
	return packer.unpack(data, return_meta=return_meta, return_corrections=return_corrections)



def save_pack(obj: SERIALIZABLE, fp: Union[TextIO, str, Path], meta: Dict[str, JSONABLE] = None,
              include_timestamp: bool = False) -> NoReturn:
	'''
	Pack (serialize) the object and store it as a json file
	
	:param obj: object to be packed
	:param fp: writable file-like object where the packed object is stored
	:param include_timestamp: include timestamp in meta information
	:return: None
	'''
	if isinstance(fp, (str, Path)):
		fp = open(str(fp), 'w')
	return json.dump(pack(obj, meta=meta, include_timestamp=include_timestamp), fp)



def load_pack(fp: Union[TextIO, str, Path], return_meta: bool = False, return_corrections: bool = False,
                allow_ancestors: bool = True) -> SERIALIZABLE:
	'''
	Loads json file of packed object and unpacks the object
	
	:param fp: writable file-like object
	:param return_meta: return the meta information stored
	:return: unpacked object from json file
	'''
	if isinstance(fp, (str, Path)):
		fp = open(str(fp), 'r')
	return unpack(json.load(fp), return_meta=return_meta, return_corrections=return_corrections,
	              allow_ancestors=allow_ancestors)



def json_pack(obj: SERIALIZABLE, meta: Dict[str, JSONABLE] = None, include_timestamp:bool = False) -> str:
	'''
	Pack object and return a json string of the serialized object
	
	:param obj: to be packed
	:param meta: any meta information to include
	:param include_timestamp: include timestamp in meta information
	:return: json string of the serialized data
	'''
	return json.dumps(pack(obj, meta=meta, include_timestamp=include_timestamp))



def json_unpack(data: str, return_meta: bool = False, return_corrections: bool = False,
                allow_ancestors: bool = True) -> SERIALIZABLE:
	'''
	Unpack json string of a packed object.
	
	:param data: json string of a packed object
	:param return_meta: return meta information
	:return: unpacked object
	'''
	return unpack(json.loads(data), return_meta=return_meta, return_corrections=return_corrections,
	              allow_ancestors=allow_ancestors)







class InitWall:
    def __init__(self, *args, _multi_inits=None, _req_args=(), _req_kwargs={}, **kwargs):
        if _multi_inits is None:
            super().__init__(*_req_args, **_req_kwargs)
        else:
            for base in _multi_inits:
                if base is None:
                    super().__init__(*_req_args.get(base, ()),
                                     **_req_kwargs.get(base, {}))

                elif isinstance(self, base):
                    super(base, self).__init__(*_req_args.get(base,()),
                                               **_req_kwargs.get(base, {}))


class Singleton(object):
    _instance = None
    
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._instance = None
    
    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance


# class InitSingleton(Singleton):
#     _instance_initialized = False
#
#     def __init__(self, *args, **kwargs):
#         if not self.__class__._instance_initialized:
#             self.__class__._instance_initialized = True
#             self.__init_singleton__(*args, **kwargs)
#
#     def __init_singleton__(self, *args, **kwargs):
#         pass










import json
import hashlib



class Hashable(object):
	'''Mixin to allow hashing'''
	
	def __hash__(self):
		return id(self)
	
	def __eq__(self, other):
		return id(self) == id(other)



def md5(s):
	if not isinstance(s, str):
		s = json.dumps(s, sort_keys=True)
	return hashlib.md5(s.encode('utf-8')).hexdigest()




from typing import Any, Union, Dict, List, Set, Tuple, NoReturn, ClassVar, TextIO, Callable

import base64
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

import hashlib

from collections import OrderedDict

from .packing import json_pack, json_unpack, SERIALIZABLE, PACKED, JSONABLE
from .errors import WrongKeyError, UnknownUserError, UnknownActionError, InsufficientPermissionsError
from cryptography.fernet import Fernet, InvalidToken, InvalidSignature
from crypt import crypt, mksalt
from getpass import getpass
# from hmac import compare_digest


_master_salt = '6FwLrxJb5mTPVwthumpackMASTERsalt'

def format_key(hsh: Union[str, bytes]) -> bytes:
	'''
	Reformat a hash (can be str or bytes)
	
	:param hsh: hash to be reformatted
	:return: a key which can be used for decrypting the data
	'''
	if isinstance(hsh, str):
		hsh = hsh.encode('latin1')
	
	kdf = PBKDF2HMAC(
		algorithm=hashes.SHA256(),
		length=32,
		salt=_master_salt.encode('latin1'),
		iterations=100000,
		backend=default_backend(),
	)
	return base64.urlsafe_b64encode(kdf.derive(hsh))

def secure_key(word: str, salt: str = None) -> str:
	'''
	Get a hash from the raw text password
	
	:param word: raw text password
	:param salt: random salt to seed the hash computation
	:return: hash of the password
	'''
	if salt is None:
		salt = _master_salt
	hsh = crypt(word, salt)
	return hsh

def prompt_password_hash(salt: str = None) -> str:
	'''
	Get the hash of a password which is entered by the user in a prompt
	
	:param salt: random salt for hash computation
	:return: hash of the entered password
	'''
	if salt is None:
		salt = _master_salt
	hsh = secure_key(getpass(), salt=salt)
	return hsh



def encrypt(data: bytes, hsh: Union[str, bytes] = None) -> bytes:
	'''
	Encrypt the data from a provided hash (or prompt for a password if no hash is provided)
	
	:param data: bytes to be encrypted
	:param hsh: hash used as key to encrypt
	:return: encrypted data bytes
	'''
	if hsh is None:
		hsh = prompt_password_hash()
	key = format_key(hsh)
	f = Fernet(key)
	return f.encrypt(data)

def decrypt(data: bytes, hsh: Union[str, bytes] = None) -> bytes:
	'''
	Decrypt data with provided hash (or prompt for a password to compute hash if no hash is provided)
	
	:param data: encrypted bytes that should be decrypted
	:param hsh: hash of a password to be used as a key to decrypt (obviously, must be the same as was used to encrypt)
	:return: the decrypted data, or a WrongKeyError if the key failed
	'''
	if hsh is None:
		hsh = prompt_password_hash()
	key = format_key(hsh)
	f = Fernet(key)
	try:
		return f.decrypt(data)
	except (InvalidToken, InvalidSignature):
		pass
	raise WrongKeyError



def secure_pack(obj: SERIALIZABLE, hsh: Union[str, bytes] = None,
                meta: Dict[str,JSONABLE] = None, include_timestamp:bool = False) -> bytes:
	'''
	Pack the object and encrypt it using the provided hash (prompt user for password if none is provided)
	
	:param obj: object to be packed
	:param hsh: hash used as key to encrypt
	:param meta: meta information to store with the packed `obj`
	:param include_timestamp: include timestamp in meta info
	:return: encrypted bytes
	'''
	data = json_pack(obj, meta=meta, include_timestamp=include_timestamp)
	data = data.encode('latin1')

	if hsh is None:
		hsh = prompt_password_hash()

	return encrypt(data, hsh)

def secure_unpack(data: bytes, hsh: Union[str, bytes] = None,
                  return_meta: bool = False) -> SERIALIZABLE:
	'''
	Decrypt `data` and unpack to recover the original object using the provided hash as a key
	
	:param data: encrypted bytes
	:param hsh: hash to be used as a key to decrypt (prompt user, if not provided)
	:param return_meta: include meta info in output
	:return: decrypted and unpacked object, possibly including the meta info
	'''
	if hsh is None:
		hsh = prompt_password_hash()
	
	data = decrypt(data, hsh).decode('latin1')
	data = json_unpack(data, return_meta=return_meta)

	return data


class Permission_Handler:  # TODO: make transactionable, actually just move to humpack
	
	def __init__(self, power_hierarchy={}, default_roles=[], god=None):
		
		self.god = god  # separate role that can't be modified by clients
		
		self.users = OrderedDict()  # user -> roles
		self.powers = OrderedDict()  # user -> power
		
		self._power_hierarchy = OrderedDict()  # role -> power
		self.roles = OrderedDict()  # role -> users
		self.update_roles(**power_hierarchy)
		
		self.default_roles = [role for role in default_roles if role in self.roles]  # for new users
		
		self.permissions = OrderedDict()  # action -> roles
		self.actions = OrderedDict()  # action -> power
	
	def _update_consistency(self, users=[], roles=[]):
		
		if not (len(users) or len(roles)):
			users = list(self.users.keys())
			roles = list(self.roles.keys())
		
		for user in users:
			for role, usrs in self.roles.items():
				if role not in self.users[user] and user in usrs:
					self.users[user].add(role)
				elif role in self.users[user] and user not in usrs:
					self.users[user].remove(role)
		
		for role in roles:
			for user, rls in self.users.items():
				if user not in self.roles[role] and role in rls:
					self.roles[role].add(user)
				elif user in self.roles[role] and role not in rls:
					self.roles[role].remove(role)
	
	def __contains__(self, user):
		return self.contains_user(user)
	
	def contains_user(self, user):
		return user in self.users or user in self.powers
	
	def contains_action(self, action):
		return action in self.actions or action in self.permissions
	
	def _update_role(self, name, power):
		if name not in self.roles:
			self.roles[name] = set()
		self._power_hierarchy[name] = power
	
	def update_role(self, role, power):
		self._update_role(role, power)
		self._update_consistency(role=[role])
	
	def update_roles(self, **roles):
		for name, power in roles.items():
			self._update_role(name, power)
		self._update_consistency(role=roles)
	
	def update_user(self, user, *roles, power=None):
		
		if power is not None:
			self.powers[user] = power
		
		if len(roles):
			if user not in self.users:
				self.users[user] = set()
			self.users[user].update(roles)
	
	def new_user(self, user, *roles, power=None):
		
		roles = set(roles)
		roles.update(self.default_roles)
		
		self.update_user(user, *roles, power=power)
	
	def set_user(self, user, *roles, power=None):
		
		if power is None and user in self.powers:
			del self.powers[user]
		
		self.update_user(user, *roles, power=power)
	
	def remove_user(self, user):
		
		if user in self.users:
			del self.users[user]
		
		if user in self.powers:
			del self.powers[user]
	
	def update_action(self, action, *roles, power=None):
		
		if power is not None:
			self.actions[action] = power
		
		if len(roles):
			if action not in self.permissions:
				self.permissions[action] = set()
			self.permissions[action].update(roles)
	
	def new_action(self, action, *roles, power=None):
		self.update_action(action, *roles, power=power)
	
	def set_action(self, action, *roles, power=None):
		
		if power is None and action in self.actions:
			del self.actions[action]
		
		self.update_action(action, *roles, power=power)
	
	def remove_action(self, action):
		
		if action in self.actions:
			del self.actions[action]
		
		if action in self.permissions:
			del self.permissions[action]
	
	def validate(self, user, action=None):
		
		if user == self.god:
			return user
		
		if self.contains_user(user):
			raise UnknownUserError(user)
		
		if action is None:
			return user
		
		if not self.contains_action(action):
			raise UnknownActionError(action)
		
		if user in self.users and action in self.permissions \
				and len(self.users[user].intersection(self.permissions[action])):
			return user
		
		if user in self.powers and action in self.actions \
				and self.powers[user] >= self.actions[action]:
			return user
		
		if user in self.users and action in self.actions:
			power = max(self._power_hierarchy.get(role, 0) for role in self.users[user])
			if power >= self.actions[action]:
				return user
		
		if user in self.powers and action in self.permissions:
			req = min(self._power_hierarchy.get(role, 0) for role in self.permissions[action])
			if self.powers[user] >= req:
				return user
		
		raise InsufficientPermissionsError(user, action)








try:
	import uuid
	from IPython.display import display_javascript, display_html
	import json
except ImportError:
	pass


class bcolors:
	BLUE = '\033[94m'
	CYAN = '\033[96m'
	GREEN = '\033[92m'
	YELLOW = '\033[93m'
	RED = '\033[91m'

	HEADER = '\033[95m'
	WARNING = '\033[93m'
	FAIL = '\033[91m'

	ENDC = '\033[0m'

	BOLD = '\033[1m'
	UNDERLINE = '\033[4m'

	color_table = {
		'blue': '\033[94m',
		'cyan': '\033[96m',
		'green': '\033[92m',
		'magenta': '\033[95m',
		'yellow': '\033[93m',
		'red': '\033[91m',

		'header': '\033[95m', # magenta
		'warning': '\033[93m', # yellow
		'fail': '\033[91m', # red
	}
	fmt_table = {
		'bold': '\033[1m',
		'underline': '\033[4m',
	}

def colorize(text, color=None, fmt=None):
	if color is None and fmt is None:
		return text
	else:
		fmt = bcolors.fmt_table.get(fmt.lower(), fmt) if isinstance(fmt, str) else ''
		color = bcolors.color_table.get(color.lower(), color) if isinstance(color, str) else ''
		return f'{color}{fmt}{text}{bcolors.ENDC}'

def printc(*args, color=None, fmt=None, sep=' ', **kwargs):
	print(colorize(sep.join(map(str, args)), color=color, fmt=fmt), **kwargs)


class render_dict(object):
	def __init__(self, json_data):
		if isinstance(json_data, dict):
			self.json_str = json.dumps(json_data)
		else:
			self.json_str = json_data
		self.uuid = str(uuid.uuid4())

	def _ipython_display_(self):
		display_html('<div id="{}" style="height: 100px; width:100%;"></div>'.format(self.uuid), raw=True)
		display_javascript("""
		require(["https://rawgit.com/caldwell/renderjson/master/renderjson.js"], function() {
		document.getElementById('%s').appendChild(renderjson(%s))
		});
		""" % (self.uuid, self.json_str), raw=True)



import sys, os
import logging

LIB_PATH = os.path.dirname(__file__)

log_levels = {
	'debug': logging.DEBUG,
	'info': logging.INFO,
	'warning': logging.WARNING,
	'error': logging.ERROR,
	'critical': logging.CRITICAL,
}

global_settings = {  # primarily for logging
	'level': 'warning',
	# 'logfile': os.path.join(os.path.dirname(LIB_PATH), 'logs', 'test.log'),
	'format': '%(levelname)s:%(name)s: %(msg)s',
	'stream': True,
	
}

# _printers = []

def set_printer_setting(level=None, format=None, formatter=None):
	
	logger = logging.getLogger()
	
	if level is not None:
		logger.setLevel(log_levels[level] if level in log_levels else level)
	
	if format is not None and formatter is None:  # default formatter
		formatter = logging.Formatter(format)
	
	if not len(logger.handlers):
		stream_handler = logging.StreamHandler()
		stream_handler.setFormatter(formatter)
		logger.addHandler(stream_handler)
	
	if formatter is not None:
		logger.handlers[0].setFormatter(formatter)
	
	return logger
	

set_printer_setting(level=global_settings['level'], format=global_settings['format'])


def get_printer(name, level=None, format=None, formatter=None,

                no_file=None, file_handler=None,
                filelevel=None, fileformatter=None, filepath=None,

                include_stream=None, stream_handler=None,
                stream_level=None, stream_formatter=None,

                ):
	'''
	Create a logger possibly writing records to both a file and stdout (aka streaming).
	For any unspecified arguments, the global settings are consulted.

	:param name:
	:param level:
	:param format:
	:param formatter:
	:param no_file:
	:param file_handler:
	:param filelevel:
	:param fileformatter:
	:param filepath:
	:param include_stream:
	:param stream_handler:
	:param stream_level:
	:param stream_formatter:
	:return:
	'''
	
	# if include_stream is None:
	# 	include_stream = global_settings['stream']
	# if no_file is None:
	# 	no_file = 'logfile' in global_settings and global_settings['logfile'] is not None
	
	# assert not no_file or include_stream, 'Not logging anywhere'
	
	logger = logging.getLogger(name)

	logger_handler = logging.StreamHandler()  # Handler for the logger
	logger.addHandler(logger_handler)
	
	# if level is None:
	# 	level = global_settings['level']
	
	if level is not None:
		logger.setLevel(log_levels[level] if level in log_levels else level)
	
	if format is not None:
		formatter = logging.Formatter(format)

		logger_handler.setFormatter(formatter)
	
	# if format is None:  # default formatter
	# 	format = global_settings['format']
	# if formatter is None:
	# 	formatter = logging.Formatter(format)
	
	if filepath is None and 'logfile' in global_settings and global_settings['logfile'] is not None:
		filepath = global_settings['logfile']
	
	if filepath is not None:
		file_handler = logging.FileHandler(filepath)
	if file_handler is not None:
		if level is not None:
			filelevel = level
		if formatter is not None:
			fileformatter = formatter
		if filelevel is not None:
			file_handler.setLevel(log_levels[filelevel] if filelevel in log_levels else filelevel)
		if fileformatter is not None:
			file_handler.setFormatter(fileformatter)
		logger.addHandler(file_handler)
	
	# if stream_handler is None and include_stream:  # create default stream_handler
	#
	# 	if stream_level is None:
	# 		stream_level = level
	#
	# 	if stream_formatter is None:
	# 		stream_formatter = formatter
	#
	# 	stream_handler = logging.StreamHandler()
	# 	stream_handler.setLevel(log_levels[stream_level] if stream_level in log_levels else stream_level)
	# 	stream_handler.setFormatter(stream_formatter)
	#
	# if stream_handler is not None:
	# 	logger.addHandler(stream_handler)
	
	# _printers.append(logger)
	
	return logger



def get_global_settings():
	return global_settings.copy()


def get_global_setting(key):
	# if key not in _global_settings:
	# 	prt.error(f'Could not find {key} in global settings')
	return global_settings.get(key, None)


def set_global_setting(key, value):
	
	# if key == 'level':
	# 	for prt in _printers:
	# 		prt.setLevel(log_levels[value] if value in log_levels else value)
	
	global_settings[key] = value



from typing import List, Dict, Tuple, Optional, Union, Any, Hashable, Sequence, Callable, Generator, Type, Iterable, \
	Iterator, IO
from pathlib import Path
from itertools import chain
from collections import OrderedDict

from .typing import unspecified_argument
from .logging import get_printer
prt = get_printer(__name__)



class ExportManager:
	_export_fmts_head = []
	_export_fmts_tail = []
	_export_fmt_types: Dict[Type, List[Type['Exporter']]] = OrderedDict()
	_export_fmt_exts: Dict[str, List[Type['Exporter']]] = OrderedDict()

	class UnknownExportData(Exception):
		def __init__(self, obj):
			super().__init__(f'{obj}')
			self.obj = obj

	class UnknownExportPath(Exception):
		def __init__(self, path):
			super().__init__(f'{str(path)}')
			self.path = path

	class UnknownExportFormat(Exception):
		def __init__(self, fmt):
			super().__init__(f'{fmt}')
			self.fmt = fmt

	class ExportFailedError(ValueError):
		def __init__(self, obj, fmts):
			super().__init__(f'{type(obj).__name__} failed using: {", ".join(fmts)}')
			self.obj = obj
			self.fmts = fmts

	class LoadFailedError(Exception):
		def __init__(self, path, fmts):
			super().__init__(f'{path} failed using: {", ".join(fmts)}')
			self.path = path
			self.fmts = fmts

	class AmbiguousLoadPathError(FileNotFoundError):
		def __init__(self, options):
			super().__init__(f'{options}')
			self.options = options

	class ExportManagerInitError(Exception):
		pass

	def __init_subclass__(cls, inherit_exporters=True, set_as_current=False, **kwargs):
		super().__init_subclass__(**kwargs)

		head, tail = [], []
		typs, exts = OrderedDict(), OrderedDict()

		if inherit_exporters:
			for base in cls.__bases__:
				if issubclass(base, ExportManager):
					head.extend(base._export_fmts_head)
					tail.extend(base._export_fmts_tail)
					for k, v in base._export_fmt_types.items():
						typs.setdefault(k, []).extend(v)
					for k, v in base._export_fmt_exts.items():
						exts.setdefault(k, []).extend(v)

		cls._export_fmts_head = head
		cls._export_fmts_tail = tail
		cls._export_fmt_types = typs
		cls._export_fmt_exts = exts

		if set_as_current:
			set_export_manager(cls)


	def __init__(self, *args, **kwargs):
		raise self.ExportManagerInitError('ExportManager should not be instantiated')


	@classmethod
	def resolve_fmt_from_obj(cls, obj: Any) -> Iterator[Type['Exporter']]:
		missing = True
		for fmt in chain(reversed(cls._export_fmt_types.get(type(obj), [])),
		                 reversed(cls._export_fmts_head),
		                 cls._export_fmts_tail):
			if fmt.validate_export_obj(obj):
				missing = False
				yield fmt
		if missing:
			raise cls.UnknownExportData(obj)

	@classmethod
	def resolve_fmt_from_path(cls, path: Path) -> Iterator[Type['Exporter']]:
		missing = True
		for fmt in chain(reversed(cls._export_fmt_exts.get(path.suffix, [])),
		                 reversed(cls._export_fmts_head),
		                 cls._export_fmts_tail):
			if fmt.validate_export_path(path):
				missing = False
				yield fmt

		if missing:
			raise cls.UnknownExportPath(path)

	@classmethod
	def resolve_fmt(cls, fmt: Union[str, Type['Exporter']]) -> Iterator[Type['Exporter']]:
		if issubclass(fmt, Exporter):
			yield fmt
		else:
			try:
				yield from cls.resolve_fmt_from_path(Path(f'null.{fmt}' if len(fmt) else 'null'))
			except cls.UnknownExportPath:
				raise cls.UnknownExportFormat(fmt)

	@classmethod
	def create_export_path(cls, name: str, *, root: Optional[Union[str, Path]] = None,
	                       fmt: Optional[str] = None) -> Path:
		if root is None:
			root = Path()
		root = Path(root)

		if fmt is None:
			return root / name

		for fmt in cls.resolve_fmt(fmt):
			return fmt._create_export_path(name, root, src=cls)


	@classmethod
	def create_load_path(cls, name: str, root: Optional[Union[str, Path]] = None):
		if root is None:
			root = Path()
		root = Path(root)

		if not root.exists():
			raise FileNotFoundError(root)
		options = list(root.glob(f'{name}*'))
		if not len(options):
			raise FileNotFoundError(root / name)
		if len(options) > 1:
			raise cls.AmbiguousLoadPathError(options)
		return options[0]

	@classmethod
	def _export_fmt(cls, fmt: Type['Exporter'], payload: Any, path: Path, **kwargs: Any) -> Path:
		return fmt._export_payload(payload, path=path, src=cls, **kwargs)

	@classmethod
	def _load_export_fmt(cls, fmt: Type['Exporter'], path: Path, **kwargs: Any) -> Path:
		return fmt._load_export(path, src=cls, **kwargs)

	@classmethod
	def export(cls, payload: Any, name: Optional[str] = None, root: Optional[Union[str, Path]] = None,
	           fmt: Optional[Union[str, Type['Exporter']]] = None, path: Optional[Union[str, Path]] = None,
	           **kwargs) -> Path:
		assert path is not None or name is not None, f'Must provide either a path or a name to export: {payload}'
		if root is not None:
			root = Path(root)

		fmts = cls.resolve_fmt_from_obj(payload) if fmt is None else cls.resolve_fmt(fmt)
		for fmt in fmts:
			dest = fmt.create_export_path(name=name, root=root, payload=payload) if path is None else Path(path)
			try:
				return cls._export_fmt(fmt, payload, dest, **kwargs)
			except fmt.ExportFailedError:
				pass

		raise cls.ExportFailedError(payload, fmts)

	@classmethod
	def load_export(cls, name: Optional[str] = None, root: Optional[Union[str, Path]] = None, *,
	                fmt: Optional[Union[str, Type['Exporter']]] = None, path: Optional[Union[str, Path]] = None,
	                **kwargs) -> Any:
		if root is not None:
			root = Path(root)

		if fmt is not None:
			fmts = cls.resolve_fmt(fmt)
		elif path is not None:
			fmts = cls.resolve_fmt_from_path(path)
		else:
			path = cls.create_load_path(name=name, root=root)
			fmts = cls.resolve_fmt_from_path(path)

		for fmt in fmts:
			dest = fmt.create_export_path(name=name, root=root) if path is None else Path(path)
			try:
				return cls._load_export_fmt(fmt, dest, **kwargs)
			except fmt.LoadFailedError:
				pass

		raise cls.LoadFailedError(path, fmts)

	@classmethod
	def register(cls, exporter: Type['Exporter'], extensions: Optional[Union[str, Sequence[str]]] = None,
	             types: Optional[Union[Type, Sequence[Type]]] = None,
	             head: Optional[bool] = None, tail: Optional[bool] = None):

		if head is None and tail is None:
			head, tail = types is None, False
		if head:
			cls._export_fmts_head.append(exporter)
		if tail:
			cls._export_fmts_tail.append(exporter)

		if extensions is not None:
			if isinstance(extensions, str):
				extensions = (extensions,)
			else:
				extensions = tuple(extensions)
			for ext in extensions:
				if ext not in cls._export_fmt_exts:
					cls._export_fmt_exts[ext] = []
				cls._export_fmt_exts[ext].append(exporter)

		if types is not None:
			if isinstance(types, type):
				types = (types,)
			else:
				types = tuple(types)
			for typ in types:
				if typ not in cls._export_fmt_types:
					cls._export_fmt_types[typ] = []
				cls._export_fmt_types[typ].append(exporter)

		if extensions is None and types is None and not head and not tail:
			prt.warning(f'Exporter {exporter} is not registered to any extensions or types')



_current_export_manager = ExportManager
def set_export_manager(manager: Type[ExportManager]) -> Type[ExportManager]:
	global _current_export_manager
	old = _current_export_manager
	_current_export_manager = manager
	return old


def export(obj, name=None, root=None, *, fmt=None, path=None, manager=None, **kwargs):
	if manager is None:
		manager = _current_export_manager
	return manager.export(obj, name=name, root=root, fmt=fmt, path=path, **kwargs)


def load_export(name=None, root=None, *, fmt=None, path=None, manager=None, **kwargs):
	if manager is None:
		manager = _current_export_manager
	return manager.load_export(name=name, root=root, fmt=fmt, path=path, **kwargs)



class Exporter:
	@classmethod
	def validate_export_obj(cls, obj: Any) -> bool:
		options = getattr(cls, '_my_export_types', None)
		return options is not None and isinstance(obj, options)

	@classmethod
	def validate_export_path(cls, path: Path) -> bool:
		suffix = path.suffix
		options = getattr(cls, '_my_export_extensions', None)
		return options is not None and len(suffix) and suffix[1:] in options

	
	def __init_subclass__(cls, extensions: Optional[Union[str, Sequence[str]]] = None,
	                      types: Optional[Union[Type, Sequence[Type]]] = None,
			              head: Optional[bool] = None, tail: Optional[bool] = None,
	                      manager: Optional[ExportManager] = None, **kwargs):
		if extensions is not None:
			extensions = (extensions,) if isinstance(extensions, str) else tuple(extensions)
			extensions = tuple(ext if ext.startswith('.') else f'.{ext}' for ext in extensions)
		if types is not None:
			types = (types,) if isinstance(types, type) else tuple(types)

		_auto_manager = False
		if manager is None:
			_auto_manager = True
			manager = _current_export_manager

		super().__init_subclass__(**kwargs)
		manager.register(cls, extensions=extensions, types=types, head=head, tail=tail)

		if extensions is not None:
			cls._my_export_extensions = extensions
		if types is not None:
			cls._my_export_types = types
		if manager is not None and not _auto_manager:
			cls._my_export_manager = manager


	@classmethod
	def load_export(cls, name: Optional[str] = None, root: Optional[Union[Path, str]] = None, *,
	                path: Optional[Union[str, Path]] = None, manager: Optional['ExportManager'] = None) -> Any:
		if manager is None:
			manager = getattr(cls, '_my_export_manager', _current_export_manager)
		return manager.load_export(name=name, root=root, path=path, fmt=cls)

	@classmethod
	def export(cls, payload, name: Optional[str] = None, root: Optional[Union[str, Path]] = None, *,
	           path: Optional[Union[str, Path]] = None, manager: Optional['ExportManager'] = None) -> Optional[Path]:
		if manager is None:
			manager = getattr(cls, '_my_export_manager', _current_export_manager)
		return manager.export(payload, name=name, root=root, path=path, fmt=cls)

	@classmethod
	def create_export_path(cls, name: str, root: Optional[Union[Path, str]], *,
	                       payload: Optional[Any] = unspecified_argument) -> Path:
		if root is None:
			root = Path()
		root = Path(root)
		options = getattr(cls, '_my_export_extensions', None)
		return root / f'{name}{"" if options is None else options[0]}'


	class LoadFailedError(ValueError): pass
	class ExportFailedError(ValueError): pass

	@staticmethod
	def _load_export(path: Union[Path, str], src: Type['ExportManager']) -> Any:
		raise NotImplementedError

	@staticmethod
	def _export_payload(payload: Any, path: Union[Path, str], src: Type['ExportManager']) -> Optional[Path]:
		raise NotImplementedError



class CollectiveExporter(Exporter):
	'''Usually braod exporters that can export multiple types of objects (eg. pickle, json, etc.)'''
	def __init_subclass__(cls, extensions=None, head=None, tail=None, **kwargs):
		super().__init_subclass__(extensions=extensions, head=head, tail=tail, **kwargs)


class SelectiveExporter(Exporter):
	'''Usually specific exporters where the produced file is specific to the object type (eg. .png, .jpg, etc.)'''
	def __init_subclass__(cls, extensions=None, types=None, **kwargs):
		super().__init_subclass__(extensions=extensions, types=types, **kwargs)


class Exportable(SelectiveExporter):
	'''Mixin for objects that can be exported with a custom export function :func:`_export_payload()`'''
	def __init_subclass__(cls, extensions=None, types=None, **kwargs):
		if types is None:
			types = cls
		super().__init_subclass__(extensions=extensions, types=types, **kwargs)

	def export(self, name: Optional[str] = None, root: Optional[Union[str, Path]] = None,
	           manager: Optional['ExportManager'] = None) -> Optional[Path]:
		return super().export(self, name=name, root=root, manager=manager)










from c3linearize import linearize
from collections import deque


def sort_by(seq, vals, reverse=False):
	return [x[0] for x in sorted(zip(seq, vals), key=lambda x: x[1], reverse=reverse)]

def resolve_order(key, *srcs, valid=None):
	'''
	All sources must either be callable or a dict-like (implementing .get())
	If no ``valid`` is provided, only outputs that are ``None`` is treated as invalid.
	
	:param key: key to be found
	:param srcs: ordered sources that may contain key
	:param valid: callable returns a bool whether or not the source output should be accepted
	:return: output corresponding to key of the first source that produced a valid output
	'''
	for src in srcs:
		val = src(key) if callable(src) else src.get(key, None)
		if (valid is None and val is not None) or (valid is not None and valid(val)):
			return val


class CycleDetectedError(Exception):
	def __init__(self, node):
		super().__init__('Cycle detected near: {}'.format(node))


# graph = {0:[1,2], 1:[3,4], 2:[4,6], 3:[7], 4:[5], 5:[], 6:[], 7:[]}
# util.toposort(0,lambda x: graph[x], depth_first=True)
# produces: [0, 1, 3, 7, 2, 4, 6, 5]
# but should be: [0, 1, 3, 7, 2, 4, 5, 6]

def graph_conv(x, g, d=None):
	if d is None:
		d = {}
	if x not in d:
		e = g(x)
		if x not in d:
			d[x] = e
		for v in e:
			graph_conv(v, g, d)
	return d


# def toposort(root, src):
# 	return linearize(src, heads=[root], order=True)[root]

def toposort(root, get_edges, ordered=True): # TODO: this is super messy and must be cleaned up
	src = graph_conv(root, get_edges)
	
	return linearize(src, heads=[root], order=ordered)[root]


def _toposort_bad(root, get_edges, depth_first=False):
	if depth_first:
		raise NotImplementedError  # not working atm
	
	order = [root]
	done = set(order)
	options = deque(get_edges(root))
	
	while len(options):
		
		next = None
		for node in options:
			success = True
			for check in options:
				if node in get_edges(check):
					success = False
					break
			if success:
				next = node
				break
		if next is None:
			raise CycleDetectedError(node)
		else:
			options.remove(next)
			if next not in done:
				order.append(next)
				done.add(next)
				if depth_first:
					options.extendleft(reversed(get_edges(next)))
				else:
					options.extend(get_edges(next))
	
	return order






import sys, os
# import torch
import numpy as np
# import torch.multiprocessing as mp
import multiprocessing as mp
import itertools
import traceback


class WorkerPool:
	@staticmethod
	def worker_loop(fn, in_q, out_q):
		while True:
			i, args = in_q.get(block=True)
			if args is None:
				break
			args, kwargs = args
			out_q.put((i, fn(*args, **kwargs)))
		out_q.put((None, None))
	
	def __init__(self, fn, n_workers=1):
		self.fn = fn
		self.n_workers = n_workers
		self.in_q = mp.Queue()
		self.out_q = mp.Queue()
		self.workers = [mp.Process(target=self.worker_loop, args=(fn, self.in_q, self.out_q))
		                for _ in range(n_workers)]
		for w in self.workers:
			w.start()
	
	def __call__(self, *args, **kwargs):
		self.in_q.put((None, (args, kwargs)))
	
	def __del__(self):
		for _ in self.workers:
			self.in_q.put((None, None))
		for w in self.workers:
			w.join()
	
	def multi_call(self, args_list, wait=True):
		N = 0
		for i, args in enumerate(args_list):
			self.in_q.put((i, args))
			N += 1
		if not wait:
			return N
		out_list = [None] * N
		for _ in range(N):
			i, out = self.get()
			out_list[i] = out
		return out_list
	
	def get(self, wait=True):
		return self.out_q.get(block=wait)
	
	def get_forever(self):
		while True:
			yield self.get()



class ExceptionWrapper(object):
	r"""Wraps an exception plus traceback to communicate across threads"""
	def __init__(self, exc_info):
		# It is important that we don't store exc_info, see
		# NOTE [ Python Traceback Reference Cycle Problem ]
		self.exc_type = exc_info[0]
		self.exc_msg = "".join(traceback.format_exception(*exc_info))

def _worker_loop(fn, private_args, in_queue, out_queue, unique_args={}, init_fn=None):
	# torch.set_num_threads(1)
	
	output_args = None
	if init_fn is not None:
		args = private_args.copy()
		args.update(unique_args)
		try:
			output_args = init_fn(**args)
		except Exception:
			print('failed')
			out_queue.put(ExceptionWrapper(sys.exc_info()))
			return
	
	while True:
		all_args = private_args.copy()
		all_args.update(unique_args)
		if output_args is not None:
			all_args.update(output_args)
		
		args = in_queue.get()
		if args is None:
			break
		try:
			shared_args, volatile_args = args
			all_args.update(shared_args)
			all_args.update(volatile_args)
			output = fn(**all_args)
		except Exception:
			out_queue.put(ExceptionWrapper(sys.exc_info()))
		else:
			out_queue.put(output)

class Farmer(object):
	'''
	Farms computation (of functions with both private and shared args) to other processes
	4 types of arguments (all of which are dicts):
	- shared_args = arguments to be sent to workers when dispatched from master process
	- private_args = arguments that all workers own privately (eg. data you dont want to pass in each dispatch)
	- unique_worker_args = arguments that each worker owns privately (eg. random seeds)
	- volatile_args = arguments generated and sent for each dispatch

	'''
	def __init__(self, fn, shared_args={}, private_args={}, unique_worker_args=None, volatile_gen=None,
	             init_fn=None, num_workers=0, timeout=20, daemon=True, waiting=None, auto_dispatch=True):
		'''

		:param fn:
		:param shared_args:
		:param private_args:
		:param unique_worker_args:
		:param volatile_gen:
		:param init_fn:
		:param num_workers:
		:param timeout:
		:param waiting:
		:param auto_dispatch:
		'''
		
		if unique_worker_args is not None:
			try:
				num_workers = len(unique_worker_args)
			except TypeError:
				pass
		
		self.num_workers = num_workers
		self.shared_args = shared_args
		self.volatile_gen = volatile_gen
		self.timeout = timeout
		self.workers = None
		
		self.in_queue = mp.Queue()
		self.outstanding = 0
		self.auto_dispatch = auto_dispatch
		
		self.dispatch_num = 1
		
		if num_workers > 0:
			if unique_worker_args is not None:  # list of dicts
				assert len(unique_worker_args) == num_workers
			else:
				unique_worker_args = [{}] * num_workers
			
			self.out_queue = mp.Queue()
			self.workers = [
				mp.Process(target=_worker_loop, daemon=daemon,
				           args=(fn, private_args, self.in_queue, self.out_queue, u, init_fn))
				for i, u in enumerate(unique_worker_args)]
			
			for w in self.workers:
				w.daemon = True  # ensure that the worker exits on process exit
				w.start()
			
			if waiting is None:
				waiting = num_workers if auto_dispatch else 0
			self._dispatch(waiting)
		
		else:
			self.fn = fn
			self.args = private_args.copy()
			if init_fn is not None:
				output_args = init_fn(**private_args)
				if output_args is not None:
					self.args.update(output_args)
	
	def _get_volatile_args(self):
		if self.volatile_gen is not None:
			return next(self.volatile_gen)
		return {}
	
	def _dispatch(self, n=1, args=None):
		if args is None:
			args = self.shared_args
		for _ in range(n):
			try:
				self.in_queue.put((args, self._get_volatile_args()))
				self.outstanding += 1
			except StopIteration:
				pass
	
	def dispatch(self, **kwargs):
		self._dispatch(self.dispatch_num, args=kwargs.update(self.shared_args))
	
	def __len__(self):
		return self.outstanding
	
	def __iter__(self):
		return self
	
	def __next__(self):
		if self.auto_dispatch:
			self._dispatch()
		if self.outstanding == 0:
			raise StopIteration
		if self.workers is None: # apply fn in this process
			args = self.args.copy()
			shared_args, volatile_args = self.in_queue.get()
			self.outstanding -= 1
			args.update(shared_args)
			args.update(volatile_args)
			return self.fn(**args)
		output = self.out_queue.get(timeout=self.timeout)
		self.outstanding -= 1
		if isinstance(output, ExceptionWrapper):
			try:
				raise output.exc_type(output.exc_msg)
			except:
				print('***ERROR: Exception of type {} occurred'.format(output.exc_type))
				raise Exception(output.exc_msg)
		# quit()
		return output
	
	def complete(self):
		return [x for x in self]
	
	def __del__(self):
		if self.workers is not None:
			for _ in self.workers:
				self.in_queue.put(None)


def make_ghost(source, execute=None):
	'''
	upon anything done to a ghost instance it will check the 'source' for the correct behavior
	and will call 'execute' with the function and args

	:param execute: should be callable with signature: execute(fn, args=[], kwargs={})
	:param source: class which contains the functionality
	:return: Ghost object which can be used to execute functions from 'source' in 'parent'
	'''
	
	if execute is None:
		def execute(fn, args=[], kwargs={}):
			return fn(*args, **kwargs)
	
	def make_listener(fn):
		def listener(*args, **kwargs):
			return execute(fn, args, kwargs)
		return listener
	
	class Ghost(object):
		def __getattr__(ignore, item):
			# print('rpl', item)
			if hasattr(source, item):  # any method in obj_type
				return make_listener(getattr(source, item))
			
			return execute(getattr, args=[item])
		
		# NOTE: use r.__len__()
		# def __len__(ignore):
		#	raise Exception('Python error: len(r) doesn\'t work on replicas, use r.__len__() instead')
		
		def __setattr__(self, key, value):
			execute(getattr(source, '__setattr__'), args=[key, value])
		
		def __delattr__(self, item):
			execute(getattr(source, '__delattr__'), args=[item])
		
		def __getitem__(ignore, item):
			return execute(getattr(source, '__getitem__'), args=[item])
		
		def __iter__(self):
			return itertools.zip_longest(*execute(getattr(source, '__iter__')))
		
		def __add__(ignore, other):
			return execute(getattr(source, '__add__'), args=[other])
	
	return Ghost()


# Init fn for any replica (used by Replicator, Parallelizer, and Cloner) - creates instance of replicated object
def _replica_init(obj_type, init_args, init_kwargs, unique_init_kwargs={}):
	init_kwargs.update(unique_init_kwargs)
	return {'obj': obj_type(*init_args, **init_kwargs)}

# Run fn for any replica (used by Replicator, Parallelizer, and Cloner) - applies function to be executed to instance
def _replica_run(obj, fn, args, kwargs, **other_args):
	try:
		return fn(obj, *args, **kwargs)
	except Exception as e:
		return e

class Replicator(object):  # see 'replicate' function below
	def __init__(self, obj_type, replicas=None, unique_init_kwargs=None, init_args=[], init_kwargs={},
	             timeout=20, collate=True):
		
		assert replicas is not None or unique_init_kwargs is not None, 'not sure how many replicas to make'
		if replicas is None:
			replicas = len(unique_init_kwargs)
		self.replicas = replicas
		
		replica_init_args = {
			'obj_type': obj_type,
			'init_args': init_args,
			'init_kwargs': init_kwargs,
		}
		
		if replicas > 0:
			if unique_init_kwargs is not None:  # list of dicts
				assert len(unique_init_kwargs) == replicas
			else:
				unique_init_kwargs = [{}] * replicas
			
			self.in_queues = np.array([mp.Queue() for _ in range(replicas)])
			self.out_queues = np.array([mp.Queue() for _ in range(replicas)])
			# self.in_queues = [mp.Queue() for _ in range(replicas)]
			# self.out_queues = [mp.Queue() for _ in range(replicas)]
			self.workers = [
				mp.Process(target=_worker_loop, args=(
					_replica_run, replica_init_args, in_q, out_q, {'unique_init_kwargs': unique}, _replica_init))
				for i, (unique, in_q, out_q) in enumerate(zip(unique_init_kwargs, self.in_queues, self.out_queues))]
			
			for w in self.workers:
				w.daemon = True  # ensure that the worker exits on process exit
				w.start()
		
		else:  # creates an invisible wrapper
			assert unique_init_kwargs is None
			self.obj = _replica_init(**replica_init_args)['obj']
		
		self.obj_type = obj_type
		self.collate = collate
		self._idx = None
		self.timeout = timeout
	
	def __len__(self):
		return self.replicas
	
	def _idx_execute(self, sel):
		
		if len(sel) == 0:
			return make_ghost(self.obj_type, self._execute)
		
		options = np.arange(self.replicas)
		# options = list(range(self.replicas))
		
		idx = []
		for s in sel:
			new = options[s]
			try:
				idx.extend(new)
			except:
				idx.append(new)
		
		def execute_idx(fn=None, args=[], kwargs={}):
			return self._execute(fn, args, kwargs, idx=idx)
		return make_ghost(self.obj_type, execute_idx)
	
	def __call__(self, *sel):
		return self._idx_execute(sel)
	
	def __getitem__(self, sel):
		if isinstance(sel, (int, slice)):
			sel = [sel]
		return self._idx_execute(sel)
	
	def _execute(self, fn=None, args=[], kwargs={}, idx=None):
		
		shared_args = {
			'args': args,
			'kwargs': kwargs,
			'fn': fn,
		}
		
		# dispatch job
		if self.replicas == 0:
			# print(fn, args, kwargs)
			return fn(self.obj, *args, **kwargs)
		
		in_queues = self.in_queues
		out_queues = self.out_queues
		
		if idx is not None:
			in_queues = in_queues[idx]
			out_queues = out_queues[idx]
		
		for in_q in in_queues:
			in_q.put((shared_args, {}))
		
		output = []
		for out_q in out_queues:
			output.append(out_q.get(timeout=self.timeout))
			if isinstance(output[-1], ExceptionWrapper):
				raise output.exc_type(output.exc_msg)
		
		# collate output
		if self.collate and isinstance(output[0], tuple):
			output = ([o[i] for o in output] for i in range(len(output[0])))
		
		return output

class Parallelizer(Replicator): # see 'replicate' function below
	# def __init__(self, *args, **kwargs):
	#	super(Parallelizer, self).__init__(*args, **kwargs)
	
	def _execute(self, fn=None, args=[], kwargs={}, idx=None):
		
		# dispatch job
		if self.replicas == 0:
			# print(fn, args, kwargs)
			return fn(self.obj, *args, **kwargs)
		
		in_queues = self.in_queues
		out_queues = self.out_queues
		
		if idx is not None:
			in_queues = in_queues[idx]
			out_queues = out_queues[idx]
		
		jlen = len(in_queues)
		
		if len(args) > 0:
			assert len(args[0]) == jlen
			args = zip(*args)
		else:
			args = [[]] * jlen
		
		if len(kwargs) > 0:
			# uncollate kwargs
			kwargs = [{k: v[i] for k, v in kwargs.items()} for i in range(jlen)]
		else:
			kwargs = [{}] * jlen
		
		for in_q, a, k in zip(in_queues, args, kwargs):
			in_q.put(({'args': a, 'kwargs': k, 'fn': fn}, {}))
		
		output = []
		for out_q in out_queues:
			output.append(out_q.get(timeout=self.timeout))
			if isinstance(output[-1], ExceptionWrapper):
				raise output.exc_type(output.exc_msg)
		
		# collate output
		if self.collate and isinstance(output[0], tuple):
			output = ([o[i] for o in output] for i in range(len(output[0])))
		
		return output

def replicate(*args, separate_args=False, ghost=False, **kwargs):
	'''
	Creates replica objects for multiprocessing. Each process contains a unique instance of the same class ('obj_type').

	There are 2 types of managers: Replicators and Parallelizers (chosen using 'separate_args')
	Replicator - all replicas take the same arguments when called
	Parallelizer - each replica takes different arguments (passed in as a list)

	Subsets of replicas may be addressed by indexing or calling the manager. If 'ghost' is true a ghost object will be
	returned in a addition to the manager which will apply anything done to that object to all replicas

	import time
	class Example:
		def f(self, i):
			time.sleep(0.5)
			print(i)
			return i+1

	replicator, ghost = replicate(Example, replicas=8, ghost=True)

	replicator(4).f(5) # executes f on replica 4
	replicator[0,2:6].f(5) # executes f on replicas 0 and 2:6
	replicator[0,0].f(5) # executes f on replica 0 twice
	replicator().f(5) # executes f on all replicas
	ghost.f(5) # executes f on all replicas

	# parallelizers have the same behavior except for each arg and kwarg a list the same length as the number of replicas to be executed must be passed in.

	:param args: for manager (check Replicator)
	:param separate_args: if true, then for each argument in the source function, a list of arguments must be passed one for each replica
	:param kwargs: for manager (check Replicator
	:return: manager, replicas (essentially a voodoo doll)
	'''
	
	Manager = Parallelizer if separate_args else Replicator
	
	manager = Manager(*args, **kwargs)
	
	if not ghost:
		return manager
	
	replicas = make_ghost(manager.obj_type, manager._execute)
	
	return manager, replicas # anything done to the replicas object will be applied in parallel to all replicas, replicator holds info about replicas


class Cloner(Farmer):
	'''
	executes any method in 'obj_type' N (default=num_workers) times with the same args using a pool of 'num_workers' workers

	unique init args can be passed to each worker, but args to all workers for each dispatch will be the same

	example:

	import time
	class Example:
		def f(self, i):
			time.sleep(0.5)
			print(i)
			return i+1

	clones = Cloner(Example, num_workers=4)

	out = clones(8).f(5)
	# takes about 1 sec and prints "5" 8 times
	# 'out' now contains [6, 6, 6, 6, 6, 6, 6, 6]

	'''
	def __init__(self, obj_type, default_N=None, init_args=[], init_kwargs={}, unique_worker_args=None,
	             num_workers=0, collate=True, timeout=20):
		
		assert num_workers is not None or unique_worker_args is not None, 'must specify how many workers to use'
		
		if unique_worker_args is not None:
			unique_worker_args = [{'unique_init_kwargs': unique} for unique in unique_worker_args]
			num_workers = len(unique_worker_args)
		
		worker_init_args = {
			'obj_type': obj_type,
			'init_args': init_args,
			'init_kwargs': init_kwargs,
		}
		
		super(Cloner, self).__init__(fn=_replica_run, init_fn=_replica_init,
		                             private_args=worker_init_args, unique_worker_args=unique_worker_args,
		                             num_workers=num_workers, auto_dispatch=False, timeout=timeout)
		self.default_N = max(num_workers, 1) if default_N is None else default_N
		self.obj_type = obj_type
		self.collate = collate
	
	# self.clones = make_ghost(self.obj_type, self._execute)
	
	def __call__(self, N=None):
		def execute_N(fn=None, args=[], kwargs={}):
			return self._execute(fn, args, kwargs, N)
		return make_ghost(self.obj_type, execute_N)
	
	def _execute(self, fn=None, args=[], kwargs={}, N=None): # execution should not change the state of the clone
		
		self.shared_args = {
			'args': args,
			'kwargs': kwargs,
			'fn': fn,
		}
		
		if N is None:
			N = self.default_N
		
		# dispatch job
		self._dispatch(N)
		
		# collect responses
		output = [out for out in self]
		
		# collate output
		if self.collate and isinstance(output[0], tuple):
			output = ([o[i] for o in output] for i in range(len(output[0])))
		
		return output


